@@ -1,3 +1,20 @@
+# Copyright 2009-2013 Justin Riley
+#
+# This file is part of StarCluster.
+#
+# StarCluster is free software: you can redistribute it and/or modify it under
+# the terms of the GNU Lesser General Public License as published by the Free
+# Software Foundation, either version 3 of the License, or (at your option) any
+# later version.
+#
+# StarCluster is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
+# details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with StarCluster. If not, see <http://www.gnu.org/licenses/>.
+
 from datetime import datetime, timedelta
 
 from starcluster import utils
@@ -29,6 +46,8 @@ class CmdSpotHistory(CmdBase):
         now = utils.datetime_tuple_to_iso(now_tup)
         delta = now_tup - timedelta(days=30)
         thirty_days_ago = utils.datetime_tuple_to_iso(delta)
+        parser.add_option("-z", "--zone", dest="zone", default=None,
+                          help="limit results to specific availability zone")
         parser.add_option("-d", "--days", dest="days_ago",
                           action="store", type="float",
                           help="provide history in the last DAYS_AGO days "
@@ -47,7 +66,7 @@ class CmdSpotHistory(CmdBase):
                           help="plot spot history using matplotlib")
 
     def execute(self, args):
-        instance_types = ', '.join(static.INSTANCE_TYPES.keys())
+        instance_types = ', '.join(sorted(static.INSTANCE_TYPES.keys()))
         if len(args) != 1:
             self.parser.error(
                 'please provide an instance type (options: %s)' %
@@ -55,7 +74,7 @@ class CmdSpotHistory(CmdBase):
         instance_type = args[0]
         if not instance_type in static.INSTANCE_TYPES:
             self.parser.error('invalid instance type. possible options: %s' %
-                              ', '.join(static.INSTANCE_TYPES))
+                              instance_types)
         start = self.opts.start_time
         end = self.opts.end_time
         if self.opts.days_ago:
@@ -64,5 +83,6 @@ class CmdSpotHistory(CmdBase):
             start = utils.datetime_tuple_to_iso(
                 now - timedelta(days=self.opts.days_ago))
         browser_cmd = self.cfg.globals.get("web_browser")
-        self.ec2.get_spot_history(instance_type, start, end, self.opts.plot,
+        self.ec2.get_spot_history(instance_type, start, end,
+                                  zone=self.opts.zone, plot=self.opts.plot,
                                   plot_web_browser=browser_cmd)
