@@ -1,3 +1,21 @@
+# Copyright 2009-2013 Justin Riley
+#
+# This file is part of StarCluster.
+#
+# StarCluster is free software: you can redistribute it and/or modify it under
+# the terms of the GNU Lesser General Public License as published by the Free
+# Software Foundation, either version 3 of the License, or (at your option) any
+# later version.
+#
+# StarCluster is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
+# details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with StarCluster. If not, see <http://www.gnu.org/licenses/>.
+
+import re
 import time
 import stat
 import base64
@@ -9,6 +27,7 @@ from starcluster import static
 from starcluster import sshutils
 from starcluster import awsutils
 from starcluster import managers
+from starcluster import userdata
 from starcluster import exception
 from starcluster.logger import log
 
@@ -17,16 +36,11 @@ class NodeManager(managers.Manager):
     """
     Manager class for Node objects
     """
-    def ssh_to_node(self, node_id, user='root', command=None):
+    def ssh_to_node(self, node_id, user='root', command=None,
+                    forward_x11=False, forward_agent=False):
         node = self.get_node(node_id, user=user)
-        if command:
-            current_user = node.ssh.get_current_user()
-            node.ssh.switch_user(user)
-            node.ssh.execute(command, silent=False)
-            node.ssh.switch_user(current_user)
-            return node.ssh.get_last_status()
-        else:
-            node.shell(user=user)
+        return node.shell(user=user, command=command, forward_x11=forward_x11,
+                          forward_agent=forward_agent)
 
     def get_node(self, node_id, user='root'):
         """Factory for Node class"""
@@ -76,6 +90,7 @@ class Node(object):
         self._ssh = None
         self._num_procs = None
         self._memory = None
+        self._user_data = None
 
     def __repr__(self):
         return '<Node: %s (%s)>' % (self.alias, self.id)
@@ -95,6 +110,21 @@ class Node(object):
                           "retrying fetching user data (tries: %s)" % (i + 1))
                 time.sleep(5)
 
+    @property
+    def user_data(self):
+        if not self._user_data:
+            try:
+                raw = self._get_user_data()
+                self._user_data = userdata.unbundle_userdata(raw)
+            except IOError, e:
+                parent_cluster = self.parent_cluster
+                if self.parent_cluster:
+                    raise exception.IncompatibleCluster(parent_cluster)
+                else:
+                    raise exception.BaseException(
+                        "Error occurred unbundling userdata: %s" % e)
+        return self._user_data
+
     @property
     def alias(self):
         """
@@ -105,25 +135,56 @@ class Node(object):
         if not self._alias:
             alias = self.tags.get('alias')
             if not alias:
-                user_data = self._get_user_data(tries=5)
-                aliases = user_data.split('|')
+                aliasestxt = self.user_data.get(static.UD_ALIASES_FNAME)
+                aliases = aliasestxt.splitlines()[2:]
                 index = self.ami_launch_index
                 try:
                     alias = aliases[index]
                 except IndexError:
-                    log.debug(
-                        "invalid user_data: %s (index: %d)" % (aliases, index))
                     alias = None
+                    log.debug("invalid aliases file in user_data:\n%s" %
+                              aliasestxt)
                 if not alias:
                     raise exception.BaseException(
                         "instance %s has no alias" % self.id)
                 self.add_tag('alias', alias)
-            name = self.tags.get('Name')
-            if not name:
+            if not self.tags.get('Name'):
                 self.add_tag('Name', alias)
             self._alias = alias
         return self._alias
 
+    def get_plugins(self):
+        plugstxt = self.user_data.get(static.UD_PLUGINS_FNAME)
+        payload = plugstxt.split('\n', 2)[2]
+        plugins_metadata = utils.decode_uncompress_load(payload)
+        plugs = []
+        for klass, args, kwargs in plugins_metadata:
+            mod_path, klass_name = klass.rsplit('.', 1)
+            try:
+                mod = __import__(mod_path, fromlist=[klass_name])
+                plug = getattr(mod, klass_name)(*args, **kwargs)
+            except SyntaxError, e:
+                raise exception.PluginSyntaxError(
+                    "Plugin %s (%s) contains a syntax error at line %s" %
+                    (klass_name, e.filename, e.lineno))
+            except ImportError, e:
+                raise exception.PluginLoadError(
+                    "Failed to import plugin %s: %s" %
+                    (klass_name, e[0]))
+            except Exception as exc:
+                log.error("Error occured:", exc_info=True)
+                raise exception.PluginLoadError(
+                    "Failed to load plugin %s with "
+                    "the following error: %s - %s" %
+                    (klass_name, exc.__class__.__name__, exc.message))
+            plugs.append(plug)
+        return plugs
+
+    def get_volumes(self):
+        volstxt = self.user_data.get(static.UD_VOLUMES_FNAME)
+        payload = volstxt.split('\n', 2)[2]
+        return utils.decode_uncompress_load(payload)
+
     def _remove_all_tags(self):
         tags = self.tags.keys()[:]
         for t in tags:
@@ -153,13 +214,10 @@ class Node(object):
 
     @property
     def parent_cluster(self):
-        cluster_tag = "--UNKNOWN--"
         try:
-            cg = self.cluster_groups[0].name
-            cluster_tag = cg.replace(static.SECURITY_GROUP_PREFIX + '-', '')
+            return self.cluster_groups[0]
         except IndexError:
             pass
-        return cluster_tag
 
     @property
     def num_processors(self):
@@ -270,7 +328,31 @@ class Node(object):
 
     @property
     def root_device_name(self):
-        return self.instance.root_device_name
+        root_dev = self.instance.root_device_name
+        bmap = self.block_device_mapping
+        if bmap and root_dev not in bmap and self.is_ebs_backed():
+            # Hack for misconfigured AMIs (e.g. CentOS 6.3 Marketplace) These
+            # AMIs have root device name set to /dev/sda1 but no /dev/sda1 in
+            # block device map - only /dev/sda. These AMIs somehow magically
+            # work so check if /dev/sda exists and return that instead to
+            # prevent detach_external_volumes() from trying to detach the root
+            # volume on these AMIs.
+            log.warn("Root device %s is not in the block device map" %
+                     root_dev)
+            log.warn("This means the AMI was registered with either "
+                     "an incorrect root device name or an incorrect block "
+                     "device mapping")
+            sda, sda1 = '/dev/sda', '/dev/sda1'
+            if root_dev == sda1:
+                log.info("Searching for possible root device: %s" % sda)
+                if sda in self.block_device_mapping:
+                    log.warn("Found '%s' - assuming its the real root device" %
+                             sda)
+                    root_dev = sda
+                else:
+                    log.warn("Device %s isn't in the block device map either" %
+                             sda)
+        return root_dev
 
     @property
     def root_device_type(self):
@@ -328,8 +410,8 @@ class Node(object):
             key = name
             if key_by_uid:
                 key = uid
-            user_map[key] = utils.struct_passwd([name, passwd, uid, gid,
-                                               gecos, home, shell])
+            user_map[key] = utils.struct_passwd([name, passwd, uid, gid, gecos,
+                                                 home, shell])
         return user_map
 
     def getgrgid(self, gid):
@@ -409,6 +491,7 @@ class Node(object):
         ssh_folder = posixpath.join(home_folder, '.ssh')
         if not self.ssh.isdir(ssh_folder):
             self.ssh.mkdir(ssh_folder)
+        self.ssh.chown(user.pw_uid, user.pw_gid, ssh_folder)
         private_key = posixpath.join(ssh_folder, 'id_rsa')
         public_key = private_key + '.pub'
         authorized_keys = posixpath.join(ssh_folder, 'authorized_keys')
@@ -475,8 +558,8 @@ class Node(object):
         for node in nodes:
             server_pkey = node.ssh.get_server_public_key()
             node_names = {}.fromkeys([node.alias, node.private_dns_name,
-                                       node.private_dns_name_short],
-                                      node.private_ip_address)
+                                      node.private_dns_name_short],
+                                     node.private_ip_address)
             node_names[node.public_dns_name] = node.ip_address
             for name, ip in node_names.items():
                 name_ip = "%s,%s" % (name, ip)
@@ -654,6 +737,41 @@ class Node(object):
             mount_map[dev] = [path, fstype, options]
         return mount_map
 
+    def get_device_map(self):
+        """
+        Returns a dictionary mapping devices->(# of blocks) based on
+        'fdisk -l' and /proc/partitions
+        """
+        dev_regex = '/dev/[A-Za-z0-9/]+'
+        r = re.compile('Disk (%s):' % dev_regex)
+        fdiskout = '\n'.join(self.ssh.execute("fdisk -l 2>/dev/null"))
+        proc_parts = '\n'.join(self.ssh.execute("cat /proc/partitions"))
+        devmap = {}
+        for dev in r.findall(fdiskout):
+            short_name = dev.replace('/dev/', '')
+            r = re.compile("(\d+)\s+%s(?:\s+|$)" % short_name)
+            devmap[dev] = int(r.findall(proc_parts)[0])
+        return devmap
+
+    def get_partition_map(self, device=None):
+        """
+        Returns a dictionary mapping partitions->(start, end, blocks, id) based
+        on 'fdisk -l'
+        """
+        fdiskout = '\n'.join(self.ssh.execute("fdisk -l %s 2>/dev/null" %
+                                              (device or '')))
+        part_regex = '/dev/[A-Za-z0-9/]+'
+        r = re.compile('(%s)\s+\*?\s+'
+                       '(\d+)(?:[-+])?\s+'
+                       '(\d+)(?:[-+])?\s+'
+                       '(\d+)(?:[-+])?\s+'
+                       '([\da-fA-F][\da-fA-F]?)' % part_regex)
+        partmap = {}
+        for match in r.findall(fdiskout):
+            part, start, end, blocks, sys_id = match
+            partmap[part] = [int(start), int(end), int(blocks), sys_id]
+        return partmap
+
     def mount_device(self, device, path):
         """
         Mount device to path
@@ -718,8 +836,9 @@ class Node(object):
         attached_vols.update(self.block_device_mapping)
         if self.is_ebs_backed():
             # exclude the root device from the list
-            if self.root_device_name in attached_vols:
-                attached_vols.pop(self.root_device_name)
+            root_dev = self.root_device_name
+            if root_dev in attached_vols:
+                attached_vols.pop(root_dev)
         return attached_vols
 
     def detach_external_volumes(self):
@@ -851,6 +970,10 @@ class Node(object):
         except exception.SSHError:
             return False
 
+    def wait(self, interval=30):
+        while not self.is_up():
+            time.sleep(interval)
+
     def is_up(self):
         if self.update() != 'running':
             return False
@@ -886,7 +1009,8 @@ class Node(object):
                                            private_key=self.key_location)
         return self._ssh
 
-    def shell(self, user=None, forward_x11=False, command=None):
+    def shell(self, user=None, forward_x11=False, forward_agent=False,
+              command=None):
         """
         Attempts to launch an interactive shell by first trying the system's
         ssh client. If the system does not have the ssh command it falls back
@@ -900,6 +1024,7 @@ class Node(object):
             label = 'instance'
             if alias == "master":
                 label = "master"
+                alias = "node"
             elif alias:
                 label = "node"
             instance_id = alias or self.id
@@ -911,6 +1036,8 @@ class Node(object):
             sshopts = '-i %s' % self.key_location
             if forward_x11:
                 sshopts += ' -Y'
+            if forward_agent:
+                sshopts += ' -A'
             ssh_cmd = static.SSH_TEMPLATE % dict(opts=sshopts, user=user,
                                                  host=self.dns_name)
             if command:
@@ -922,10 +1049,13 @@ class Node(object):
             log.debug("Using Pure-Python SSH client")
             if forward_x11:
                 log.warn("X11 Forwarding not available in Python SSH client")
+            if forward_agent:
+                log.warn("Authentication agent forwarding not available in " +
+                         "Python SSH client")
             if command:
                 orig_user = self.ssh.get_current_user()
                 self.ssh.switch_user(user)
-                self.ssh.execute(command, silent=False, source_profile=True)
+                self.ssh.execute(command, silent=False)
                 self.ssh.switch_user(orig_user)
                 return self.ssh.get_last_status()
             self.ssh.interactive_shell(user=user)
@@ -955,6 +1085,46 @@ class Node(object):
         """
         self.apt_command('install %s' % pkgs)
 
+    def yum_command(self, cmd):
+        """
+        Run a yum command with all necessary options for non-interactive use.
+        "-d 0 -e 0 -y"
+        """
+        yum_opts = ['-d', '0', '-e', '0', '-y']
+        cmd = "yum " + " ".join(yum_opts) + " " + cmd
+        self.ssh.execute(cmd)
+
+    def yum_install(self, pkgs):
+        """
+        Install a set of packages via yum.
+
+        pkgs is a string that contains one or more packages separated by a
+        space
+        """
+        self.yum_command('install %s' % pkgs)
+
+    @property
+    def package_provider(self):
+        """
+        In order to determine which packaging system to use, check to see if
+        /usr/bin/apt exists on the node, and use apt if it exists. Otherwise
+        test to see if /usr/bin/yum exists and use that.
+        """
+        if self.ssh.isfile('/usr/bin/apt'):
+            return "apt"
+        elif self.ssh.isfile('/usr/bin/yum'):
+            return "yum"
+
+    def package_install(self, pkgs):
+        """
+        Provides a declarative install packages on systems, regardless
+        of the system's packging type (apt/yum).
+        """
+        if self.package_provider == "apt":
+            self.apt_install(pkgs)
+        elif self.package_provider == "yum":
+            self.yum_install(pkgs)
+
     def __del__(self):
         if self._ssh:
             self._ssh.close()
