@@ -1,25 +1,35 @@
-"""
-ssh.py
-Friendly Python SSH2 interface.
-From http://commandline.org.uk/code/
-License: LGPL
-modified by justin riley (justin.t.riley@gmail.com)
-"""
+# Copyright 2009-2013 Justin Riley
+#
+# This file is part of StarCluster.
+#
+# StarCluster is free software: you can redistribute it and/or modify it under
+# the terms of the GNU Lesser General Public License as published by the Free
+# Software Foundation, either version 3 of the License, or (at your option) any
+# later version.
+#
+# StarCluster is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
+# details.
+#
+# You should have received a copy of the GNU Lesser General Public License
+# along with StarCluster. If not, see <http://www.gnu.org/licenses/>.
 
 import os
 import re
 import sys
 import stat
 import glob
+import atexit
 import string
 import socket
 import fnmatch
 import hashlib
 import posixpath
 
-import ssh
-from ssh import util
-from ssh import RSAKey
+import paramiko
+from paramiko import util
+from paramiko import RSAKey
 from pyasn1.codec.der import encoder
 from pyasn1.type import univ
 
@@ -69,6 +79,7 @@ class SSHClient(object):
             raise exception.SSHNoCredentialsError()
         self._glob = SSHGlob(self)
         self.__last_status = None
+        atexit.register(self.close)
 
     def load_private_key(self, private_key, private_key_pass=None):
         # Use Private Key.
@@ -97,16 +108,16 @@ class SSHClient(object):
                                                                    username))
         try:
             sock = self._get_socket(host, port)
-            transport = ssh.Transport(sock)
+            transport = paramiko.Transport(sock)
             transport.banner_timeout = timeout
         except socket.error:
             raise exception.SSHConnectionError(host, port)
         # Authenticate the transport.
         try:
             transport.connect(username=username, pkey=pkey, password=password)
-        except ssh.AuthenticationException:
+        except paramiko.AuthenticationException:
             raise exception.SSHAuthException(username, host)
-        except ssh.SSHException, e:
+        except paramiko.SSHException, e:
             msg = e.args[0]
             raise exception.SSHError(msg)
         except socket.error:
@@ -119,7 +130,7 @@ class SSHClient(object):
         self._transport = transport
         try:
             assert self.sftp is not None
-        except ssh.SFTPError, e:
+        except paramiko.SFTPError, e:
             if 'Garbage packet received' in e:
                 log.debug("Garbage packet received", exc_info=True)
                 raise exception.SSHAccessDeniedViaAuthKeys(username)
@@ -162,21 +173,21 @@ class SSHClient(object):
     def _load_rsa_key(self, private_key, private_key_pass=None):
         private_key_file = os.path.expanduser(private_key)
         try:
-            rsa_key = ssh.RSAKey.from_private_key_file(private_key_file,
-                                                       private_key_pass)
+            rsa_key = paramiko.RSAKey.from_private_key_file(private_key_file,
+                                                            private_key_pass)
             log.debug("Using private key %s (rsa)" % private_key)
             return rsa_key
-        except ssh.SSHException:
+        except paramiko.SSHException:
             log.error('invalid rsa key or passphrase specified')
 
     def _load_dsa_key(self, private_key, private_key_pass=None):
         private_key_file = os.path.expanduser(private_key)
         try:
-            dsa_key = ssh.DSSKey.from_private_key_file(private_key_file,
-                                                       private_key_pass)
+            dsa_key = paramiko.DSSKey.from_private_key_file(private_key_file,
+                                                            private_key_pass)
             log.info("Using private key %s (dsa)" % private_key)
             return dsa_key
-        except ssh.SSHException:
+        except paramiko.SSHException:
             log.error('invalid dsa key or passphrase specified')
 
     @property
@@ -184,30 +195,31 @@ class SSHClient(object):
         """Establish the SFTP connection."""
         if not self._sftp or self._sftp.sock.closed:
             log.debug("creating sftp connection")
-            self._sftp = ssh.SFTPClient.from_transport(self.transport)
+            self._sftp = paramiko.SFTPClient.from_transport(self.transport)
         return self._sftp
 
     @property
     def scp(self):
         """Initialize the SCP client."""
-        if not self._scp:
+        if not self._scp or not self._scp.transport.is_active():
             log.debug("creating scp connection")
             self._scp = scp.SCPClient(self.transport,
                                       progress=self._file_transfer_progress)
         return self._scp
 
     def generate_rsa_key(self):
-        return ssh.RSAKey.generate(2048)
+        return paramiko.RSAKey.generate(2048)
 
     def get_public_key(self, key):
         return ' '.join([key.get_name(), key.get_base64()])
 
     def load_remote_rsa_key(self, remote_filename):
         """
-        Returns ssh.RSAKey object for an RSA key located on the remote machine
+        Returns paramiko.RSAKey object for an RSA key located on the remote
+        machine
         """
         rfile = self.remote_file(remote_filename, 'r')
-        key = ssh.RSAKey(file_obj=rfile)
+        key = paramiko.RSAKey(file_obj=rfile)
         rfile.close()
         return key
 
@@ -319,27 +331,23 @@ class SSHClient(object):
         except IOError:
             return False
 
-    def chown(self, uid, gid, remote_file):
+    def chown(self, uid, gid, remote_path):
         """
-        Apply permissions (mode) to remote_file
+        Set user (uid) and group (gid) owner for remote_path
         """
-        f = self.remote_file(remote_file, 'r')
-        f.chown(uid, gid, remote_file)
-        f.close()
+        return self.sftp.chown(remote_path, uid, gid)
 
-    def chmod(self, mode, remote_file):
+    def chmod(self, mode, remote_path):
         """
-        Apply permissions (mode) to remote_file
+        Apply permissions (mode) to remote_path
         """
-        f = self.remote_file(remote_file, 'r')
-        f.chmod(mode)
-        f.close()
+        return self.sftp.chmod(remote_path, mode)
 
     def ls(self, path):
         """
         Return a list containing the names of the entries in the remote path.
         """
-        return [os.path.join(path, f) for f in self.sftp.listdir(path)]
+        return [posixpath.join(path, f) for f in self.sftp.listdir(path)]
 
     def glob(self, pattern):
         return self._glob.glob(pattern)
@@ -442,7 +450,7 @@ class SSHClient(object):
                 break
         self.scp.put(localpaths, remote_path=remotepath, recursive=recursive)
 
-    def execute_async(self, command, source_profile=False):
+    def execute_async(self, command, source_profile=True):
         """
         Executes a remote command so that it continues running even after this
         SSH connection closes. The remote process will be put into the
@@ -455,7 +463,7 @@ class SSHClient(object):
     def get_last_status(self):
         return self.__last_status
 
-    def get_status(self, command, source_profile=False):
+    def get_status(self, command, source_profile=True):
         """
         Execute a remote command and return the exit status
         """
@@ -488,7 +496,7 @@ class SSHClient(object):
                     print line,
             for line in stderr.readlines():
                 output.append(line)
-                print line
+                print line,
         if only_printable:
             output = map(lambda line: ''.join(c for c in line if c in
                                               string.printable), output)
@@ -497,18 +505,18 @@ class SSHClient(object):
 
     def execute(self, command, silent=True, only_printable=False,
                 ignore_exit_status=False, log_output=True, detach=False,
-                source_profile=False, raise_on_failure=False):
+                source_profile=True, raise_on_failure=True):
         """
         Execute a remote command and return stdout/stderr
 
         NOTE: this function blocks until the process finishes
 
         kwargs:
-        silent - do not log output to console
+        silent - don't print the command's output to the console
         only_printable - filter the command's output to allow only printable
-                        characters
+                         characters
         ignore_exit_status - don't warn about non-zero exit status
-        log_output - log output to debug file
+        log_output - log all remote output to the debug file
         detach - detach the remote process so that it continues to run even
                  after the SSH connection closes (does NOT return output or
                  check for non-zero exit status if detach=True)
@@ -527,22 +535,33 @@ class SSHClient(object):
             return
         if source_profile:
             command = "source /etc/profile && %s" % command
+        log.debug("executing remote command: %s" % command)
         channel.exec_command(command)
         output = self._get_output(channel, silent=silent,
                                   only_printable=only_printable)
         exit_status = channel.recv_exit_status()
         self.__last_status = exit_status
+        out_str = '\n'.join(output)
         if exit_status != 0:
-            msg = "command '%s' failed with status %d" % (command, exit_status)
-            if not ignore_exit_status:
-                log.error(msg)
+            msg = "remote command '%s' failed with status %d"
+            msg %= (command, exit_status)
+            if log_output:
+                msg += ":\n%s" % out_str
             else:
-                log.debug(msg)
-        if log_output:
-            for line in output:
-                log.debug(line.strip())
-        if exit_status != 0 and raise_on_failure:
-            raise exception.SSHError(msg)
+                msg += " (no output log requested)"
+            if not ignore_exit_status:
+                if raise_on_failure:
+                    raise exception.RemoteCommandFailed(
+                        msg, command, exit_status, out_str)
+                else:
+                    log.error(msg)
+            else:
+                log.debug("(ignored) " + msg)
+        else:
+            if log_output:
+                log.debug("output of '%s':\n%s" % (command, out_str))
+            else:
+                log.debug("output of '%s' has been hidden" % command)
         return output
 
     def has_required(self, progs):
@@ -611,18 +630,13 @@ class SSHClient(object):
     def interactive_shell(self, user='root'):
         orig_user = self.get_current_user()
         self.switch_user(user)
-        try:
-            chan = self._invoke_shell()
-            log.info('Starting Pure-Python SSH shell...')
-            if HAS_TERMIOS:
-                self._posix_shell(chan)
-            else:
-                self._windows_shell(chan)
-            chan.close()
-        except Exception, e:
-            import traceback
-            print '*** Caught exception: %s: %s' % (e.__class__, e)
-            traceback.print_exc()
+        chan = self._invoke_shell()
+        log.info('Starting Pure-Python SSH shell...')
+        if HAS_TERMIOS:
+            self._posix_shell(chan)
+        else:
+            self._windows_shell(chan)
+        chan.close()
         self.switch_user(orig_user)
 
     def _posix_shell(self, chan):
@@ -754,7 +768,7 @@ class SSHGlob(object):
             #dirname = unicode(dirname, encoding)
             dirname = unicode(dirname, 'UTF-8')
         try:
-            names = [os.path.basename(n) for n in self.ssh.ls(dirname)]
+            names = [posixpath.basename(n) for n in self.ssh.ls(dirname)]
         except os.error:
             return []
         if pattern[0] != '.':
@@ -796,7 +810,7 @@ def get_private_rsa_fingerprint(key_location):
     """
     try:
         k = RSAKey.from_private_key_file(key_location)
-    except ssh.SSHException:
+    except paramiko.SSHException:
         raise exception.SSHError("Invalid RSA private key file: %s" %
                                  key_location)
     params = dict(invq=util.mod_inverse(k.q, k.p), dp=k.d % (k.p - 1),
@@ -811,7 +825,7 @@ def get_private_rsa_fingerprint(key_location):
 def get_public_rsa_fingerprint(pubkey_location):
     try:
         k = RSAKey.from_private_key_file(pubkey_location)
-    except ssh.SSHException:
+    except paramiko.SSHException:
         raise exception.SSHError("Invalid RSA private key file: %s" %
                                  pubkey_location)
     md5digest = hashlib.md5(str(k)).hexdigest()
